@@ -1,6 +1,6 @@
 # -*- coding: latin-1 -*-
 #
-# Copyright 2011-2017 Ghent University
+# Copyright 2011-2020 Ghent University
 #
 # This file is part of vsc-install,
 # originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
@@ -30,16 +30,24 @@ Shared module for vsc software setup
 @author: Stijn De Weirdt (Ghent University)
 @author: Andy Georges (Ghent University)
 """
-import __builtin__
+
+from __future__ import print_function
+import sys
+
+if sys.version_info < (3, 0):
+    import __builtin__
+else:
+    import builtins as __builtin__  # make builtins accessible via same way as in Python 3
+
 import glob
 import hashlib
 import inspect
 import json
 import os
 import shutil
-import sys
 import re
 
+import setuptools
 import setuptools.command.test
 
 from distutils import log  # also for setuptools
@@ -109,7 +117,7 @@ if log.Log.__name__ != 'NewLog':
             try:
                 return self._orig_log(self, level, newmsg, args)
             except Exception:
-                print newmsg % args
+                print(newmsg % args)
 
     log.Log = NewLog
     log._global_log = NewLog()
@@ -130,7 +138,10 @@ kw = ('Kenneth Waegeman', 'Kenneth.Waegeman@UGent.be')
 lm = ('Luis Fernando Munoz Meji?as', 'luis.munoz@ugent.be')
 sdw = ('Stijn De Weirdt', 'stijn.deweirdt@ugent.be')
 wdp = ('Wouter Depypere', 'wouter.depypere@ugent.be')
-wp = ('Ward Poelmans', 'Ward.Poelmans@UGent.be')
+wp = ('Ward Poelmans', 'ward.poelmans@vub.be')
+sm = ('Samuel Moors', 'samuel.moors@vub.be')
+bh = ('Balazs Hajgato', 'Balazs.Hajgato@UGent.be')
+ad = ('Alex Domingo', 'alex.domingo.toro@vub.be')
 
 # Regexp used to remove suffixes from scripts when installing(/packaging)
 REGEXP_REMOVE_SUFFIX = re.compile(r'(\.(?:py|sh|pl))$')
@@ -148,9 +159,10 @@ URL_GHUGENT_HPCUGENT = 'https://github.ugent.be/hpcugent/%(name)s'
 
 RELOAD_VSC_MODS = False
 
-VERSION = '0.10.25'
+VERSION = '0.15.3'
 
 log.info('This is (based on) vsc.install.shared_setup %s' % VERSION)
+log.info('(using setuptools version %s located at %s)' % (setuptools.__version__, setuptools.__file__))
 
 # list of non-vsc packages that do not need python- prefix for correct rpm dependencies
 # vsc packages should be handled with clusterbuildrpm
@@ -159,9 +171,10 @@ NO_PREFIX_PYTHON_BDIST_RPM = ['pbs_python']
 
 # Hardcode map of python dependency prefix to their rpm python- flavour prefix
 PYTHON_BDIST_RPM_PREFIX_MAP = {
-    'pycrypto': 'python-crypto',
+    'pycrypto': 'python%s-crypto',
 }
 
+SHEBANG_BIN_BASH = "#!/bin/bash"
 SHEBANG_ENV_PYTHON = '#!/usr/bin/env python'
 SHEBANG_NOENV_PYTHON = '#!/usr/bin/python-noenv'
 SHEBANG_PYTHON_E = '#!/usr/bin/python -E'
@@ -316,8 +329,8 @@ class vsc_setup(object):
             ],
             'url': [
                 r'^Home-page:\s*(.*?)\s*$',
-                r'^\s*url\s*=\s*((?:https?|ssh).*?github.*?[:/]hpcugent/.*?)\.git\s*$',
-                r'^\s*url\s*=\s*(git[:@].*?github.*?[:/]hpcugent/.*?)(?:\.git)?\s*$',
+                r'^\s*url\s*=\s*((?:https?|ssh).*?github.*?[:/](?:hpcugent|sisc-hpc)/.*?)\.git\s*$',
+                r'^\s*url\s*=\s*(git[:@].*?github.*?[:/](?:hpcugent|sisc-hpc)/.*?)(?:\.git)?\s*$',
             ],
             'download_url': [
                 r'^Download-URL:\s*(.*?)\s*$',
@@ -327,7 +340,7 @@ class vsc_setup(object):
         res = {}
         for name, patterns in all_patterns.items():
             for pat in patterns:
-                reg = re.search(pat, txt, re.M)
+                reg = re.search(pat, txt[:10240], re.M)
                 if reg:
                     res[name] = reg.group(1)
                     log.info('found match %s %s in %s' % (name, res[name], filename))
@@ -337,15 +350,17 @@ class vsc_setup(object):
         reg = re.search(r'^git@(.*?):(.*)$', res.get('url', ''))
         if reg:
             res['url'] = "https://%s/%s" % (reg.group(1), reg.group(2))
+            log.info('reg found: %s', reg.groups())
             self.private_repo = True
 
         if 'url' not in res:
-            raise KeyError("Missing url in git config %s. (Missing mandatory hpcugent (upstream) remote?)" % (res))
+            raise KeyError("Missing url in git config %s. (Missing mandatory hpcugent or sisc-hpc remote?)" % (res))
 
         # handle git://server/user/project
         reg = re.search(r'^(git|ssh)://', res.get('url', ''))
         if reg:
             res['url'] = "https://%s" % res['url'][len(reg.group(0)):]
+            log.info('reg found: %s', reg.groups())
             self.private_repo = True
 
         if 'download_url' not in res:
@@ -359,13 +374,15 @@ class vsc_setup(object):
         if len(res) != 3:
             raise Exception("Cannot determine name, url and download url from filename %s: got %s" % (filename, res))
         else:
+            keepers = {}
             for name, value in res.items():
                 if value is None:
                     log.info('Removing None %s' % name)
-                    res.pop(name)
+                else:
+                    keepers[name] = value
 
-            log.info('get_name_url returns %s' % res)
-            return res
+            log.info('get_name_url returns %s' % keepers)
+            return keepers
 
     def rel_gitignore(self, paths, base_dir=None):
         """
@@ -424,7 +441,7 @@ class vsc_setup(object):
                 # Force vsc shared packages/namespace
                 if '__init__.py' in files and (package == 'vsc' or package.startswith('vsc.')):
                     init = open(os.path.join(root, '__init__.py')).read()
-                    if not re.search(r'^import\s+pkg_resources\npkg_resources.declare_namespace\(__name__\)$',
+                    if not re.search(r'^import\s+pkg_resources\n{1,3}pkg_resources.declare_namespace\(__name__\)$',
                                      init, re.M):
                         raise Exception(('vsc namespace packages do not allow non-shared namespace in dir %s.'
                                          'Fix with pkg_resources.declare_namespace') % root)
@@ -618,6 +635,8 @@ class vsc_setup(object):
         few extra files we need to add for installation purposes.
         """
 
+
+        # pylint: disable=arguments-differ
         def finalize_options(self, *args, **kwargs):
             """Handle missing lib dir for scripts-only packages"""
             # the egginfo data will be deleted as part of the cleanup
@@ -692,9 +711,9 @@ class vsc_setup(object):
             log.info("vsc_bdist_rpm = %s" % (self.__dict__))
             klass = _fvs('vsc_bdist_rpm egg_info')
             # changed to allow file removal
-            klass.SHARED_TARGET['cmdclass']['egg_info'] = klass.vsc_bdist_rpm_egg_info
+            self.distribution.cmdclass['egg_info'] = klass.vsc_bdist_rpm_egg_info
             # changed to allow modification of shebangs
-            klass.SHARED_TARGET['cmdclass']['sdist'] = klass.vsc_sdist_rpm
+            self.distribution.cmdclass['sdist'] = klass.vsc_sdist_rpm
             self.run_command('egg_info')  # ensure distro name is up-to-date
             orig_bdist_rpm.run(self)
 
@@ -1043,7 +1062,7 @@ class vsc_setup(object):
 
         def _print(self, cmd):
             """Print is evil, cmd is list"""
-            print ' '.join(cmd)
+            print(' '.join(cmd))
 
         def git_tag(self):
             """Tag the version in git"""
@@ -1103,15 +1122,21 @@ class vsc_setup(object):
             log.info('Register with pypi')
             # do actually do this, use self.run_command()
             # you can only upload what you just created
-            self._print(['# Run command below to register with pypi (testpypi %s)' % self.testpypi])
-            self._print(setup + ['register'] + test + ['sdist', 'upload'] + test)
+            self._print(['# Run commands below to upload to PyPI (testpypi %s)' % self.testpypi])
+            self._print(setup + ['register'] + test + ['sdist'])
+            self._print(['twine', 'upload', '--verbose', 'dist/%s.tar.gz' % self.distribution.get_fullname()])
 
         def run(self):
             """Print list of thinigs to do"""
             fullname = self.distribution.get_fullname()
 
             url = self.distribution.get_url()
-            gh_reg = re.search(r'^.*?://([^/]*github[^/]*)/', url)
+            # this is a regex with catastrophic backtracking, so limit the length of url. This takes 10 secs
+            # to process on 'a://a' + 'github' *12721 + 'it'
+            # thx to James Davis (https://github.com/davisjam) for pointing this out.
+            # this regex tries to parse the domain in the url to differentiate between https://github.com/ and
+            # private enterprise installs of github e.g. http://github.example.com/
+            gh_reg = re.search(r'^.*?://([^/]*github[^/]*)/', url[:1024])
 
             log.info("Release commands to perform for %s" % fullname)
             if gh_reg:
@@ -1165,7 +1190,7 @@ class vsc_setup(object):
         """
         Transforms name into a sensible string for use in setup.cfg.
 
-        enviroment variable VSC_RPM_PYTHON is set to 1 and either
+        environment variable VSC_RPM_PYTHON is set to 1,2 or 3 and either
             name starts with key from PYTHON_BDIST_RPM_PREFIX_MAP
                 new name starts with value
             python- is prefixed in case of
@@ -1173,29 +1198,50 @@ class vsc_setup(object):
                 name starts with 'vsc'
                 and name does not start with python-
         """
-        if isinstance(name, basestring):
 
-            if os.environ.get('VSC_RPM_PYTHON', 'NOT_ONE') == '1':
+        if isinstance(name, (list, tuple)):
+            klass = _fvs('sanitize')
+            return ",".join([klass.sanitize(r) for r in name])
+
+        else:
+            pyversuff = os.environ.get('VSC_RPM_PYTHON', None)
+            if pyversuff in ("1", "2", "3"):
+                # enable VSC-style naming for Python packages: use 'python2-*' or 'python3-*',
+                # unless '1' is used as value for $VSC_RPM_PYTHON, then use 'python-*' for legacy behaviour
+                if pyversuff == '1':
+                    pyversuff = ''
+
                 # hardcoded prefix map
                 for pydep, rpmname in PYTHON_BDIST_RPM_PREFIX_MAP.items():
                     if name.startswith(pydep):
-                        return rpmname+name[len(pydep):]
+                        newname = (rpmname+name[len(pydep):]) % pyversuff
+                        log.debug("new sanitized name %s from map (old %s)", newname, name)
+                        return newname
 
                 # more sensible map
-                p_p = (not ([x for x in NO_PREFIX_PYTHON_BDIST_RPM if name.startswith(x)] or
-                       name.startswith('python-')) or name.startswith('vsc'))
+                is_python_pkg = (not ([x for x in NO_PREFIX_PYTHON_BDIST_RPM if name.startswith(x)] or
+                                      name.startswith('python-') or name.startswith('python%s-' % pyversuff))
+                                 or name.startswith('vsc'))
+
+                if is_python_pkg:
+                    newname = 'python%s-%s' % (pyversuff, name)
+                    log.debug("new sanitized name %s (old %s)", newname, name)
+                    return newname
 
-                if p_p:
-                    name = 'python-%s' % name
             return name
-        else:
-            klass = _fvs('sanitize')
-            return ",".join([klass.sanitize(r) for r in name])
+
+
+
 
     @staticmethod
     def get_md5sum(filename):
         """Use this function to compute the md5sum in the KNOWN_LICENSES hash"""
-        return hashlib.md5(open(filename).read()).hexdigest()
+        hasher = hashlib.md5()
+        with open(filename, "rb") as fh:
+            for chunk in iter(lambda: fh.read(4096), b""):
+                hasher.update(chunk)
+        return hasher.hexdigest()
+
 
     def get_license(self, license_name=None):
         """
@@ -1243,21 +1289,17 @@ class vsc_setup(object):
         new_target = {}
         new_target.update(vsc_setup_klass.SHARED_TARGET)
 
-        if sys.version_info < (2, 7):
-            # py26 support dropped in 0.8, and the old versions don't detect enough
-            log.info('no prospector support in py26 (or older)')
-        else:
-            log.info('adding prospector to tests_require')
-            tests_requires = new_target.setdefault('tests_require', [])
-            tests_requires.extend(['prospector >= 0.12.1', 'pylint < 1.6.0'])
-
         # update the cmdclass with ones from vsc_setup_klass
         # cannot do this in one go, when SHARED_TARGET is defined, vsc_setup doesn't exist yet
-        for name, klass in new_target['cmdclass'].items():
+        keepers = new_target['cmdclass'].copy()
+        for name in new_target['cmdclass']:
+            klass = new_target['cmdclass'][name]
             try:
-                new_target['cmdclass'][name] = getattr(vsc_setup_klass, klass.__name__)
+                keepers[name] = getattr(vsc_setup_klass, klass.__name__)
             except AttributeError:
-                del new_target['cmdclass'][name]
+                del keepers[name]
+                log.info("Not including new_target['cmdclass']['%s']" % name)
+        new_target['cmdclass'] = keepers
 
         # prepare classifiers
         classifiers = new_target.setdefault('classifiers', [])
@@ -1268,10 +1310,13 @@ class vsc_setup(object):
         new_target['license'] = lic_name
         classifiers.append(lic_classifier)
 
-        if 'name' not in target:
-            log.info('No name defined, trying to determine it')
+        # set name, url, download_url (skip name if it was specified)
+        update = self.get_name_url(version=target['version'], license_name=lic_name)
+        if 'name' in target:
+            log.info('Name defined, not using auto determined name')
             # sets name / url and download_url
-            target.update(self.get_name_url(version=target['version'], license_name=lic_name))
+            del update['name']
+        target.update(update)
 
         if urltemplate:
             new_target['url'] = urltemplate % target
@@ -1310,6 +1355,20 @@ class vsc_setup(object):
             new_target['description'] = descr  # summary in PKG-INFO
             new_target['long_description'] = readmetxt  # description in PKG-INFO
 
+            readme_ext = os.path.splitext(readme)[-1]
+            # see https://packaging.python.org/guides/making-a-pypi-friendly-readme/
+            readme_content_types = {
+                '.md': 'text/markdown',
+                '.rst': 'text/x-rst',
+                '.txt': 'text/plain',
+                # fallback in case README file has no extension
+                '': 'text/plain',
+            }
+            if readme_ext in readme_content_types:
+                new_target['long_description_content_type'] = readme_content_types[readme_ext]
+            else:
+                raise Exception("Failed to derive content type for README file '%s' based on extension" % readme)
+
         vsc_scripts = target.pop('vsc_scripts', True)
         if vsc_scripts:
             candidates = self.generate_scripts()
@@ -1336,7 +1395,7 @@ class vsc_setup(object):
                     log.error("%s of config %s needs to be a list (not tuple or string)" % (k, target['name']))
                     sys.exit(1)
                 new_target[k] = ";".join([x[0] for x in v])
-                new_target["%s_email" % k] = ";".join([x[1] for x in v])
+                new_target["%s_email" % k] = ", ".join([x[1] for x in v])
             else:
                 if isinstance(v, dict):
                     # eg command_class
@@ -1355,33 +1414,42 @@ class vsc_setup(object):
         else:
             log.info('adding prospector to tests_require')
             tests_requires = new_target.setdefault('tests_require', [])
-            tests_requires.extend(['prospector >= 0.12.1', 'pylint < 1.6.0'])
+            # Python 2.x support was removed in pydocstyle 4.0, so stick to latest release before 4.0
+            tests_requires.append('pydocstyle < 4.0')
+            # fix from https://github.com/PyCQA/prospector/pull/323 required to avoid infinite recursion
+            tests_requires.append('prospector >= 1.1.6.4')
             new_target['tests_require'] = tests_requires
 
         if self.private_repo:
             urls = [
+                ('github.com', 'git+https://'),
                 ('github.ugent.be', 'git+ssh://git@'),
                 ('github.com', 'git+ssh://git@'),
-                ('github.com', 'git+https://'),
             ]
         else:
             urls = [('github.com', 'git+https://')]
+
         for dependency in set(new_target['install_requires'] + new_target['setup_requires'] +
             new_target['tests_require']):
+
+            # see https://docs.python.org/3/reference/expressions.html#comparisons
+            split_re = re.compile(r'(\s)?([<>]=?|[=!]=)(\s)?')
+            check_whitespace = split_re.search(dependency)
+
+            if check_whitespace and (check_whitespace.group(1) is None or check_whitespace.group(3) is None):
+                raise ValueError("Missing spaces around comparison operator in '%s'" % dependency)
+
             if dependency.startswith('vsc'):
-                dep = dependency.split(' ')[0]
-                depversion = ''
-                for comp in ['=', '<']:
-                    try:
-                        depversion = "-" + dependency.split(comp)[1].strip()
-                    except IndexError:
-                        pass
-                    for url, git_scheme in urls:
-                        new_target['dependency_links'] += [''.join([git_scheme, url, '/hpcugent/', dep, '.git#egg=',
-                                                           dep, depversion])]
+                dep_name = split_re.split(dependency)[0]
+                dep_name_version = split_re.sub('-', dependency)
+                # if you specify any kind of version on a dependency, the dependency_links also needs a version or
+                # else it's ignored: https://setuptools.readthedocs.io/en/latest/setuptools.html#id14
+                for url, git_scheme in urls:
+                    new_target['dependency_links'] += [''.join([git_scheme, url, '/hpcugent/', dep_name, '.git#egg=',
+                                                                dep_name_version])]
 
         log.debug("New target = %s" % (new_target))
-        print new_target
+        print(new_target)
         return new_target
 
     @staticmethod
@@ -1406,8 +1474,8 @@ class vsc_setup(object):
 
         try:
             setup_cfg = open('setup.cfg', 'w')  # and truncate
-        except (IOError, OSError), err:
-            print "Cannot create setup.cfg for target %s: %s" % (target['name'], err)
+        except (IOError, OSError) as err:
+            print("Cannot create setup.cfg for target %s: %s" % (target['name'], err))
             sys.exit(1)
 
         klass = _fvs('build_setup_cfg_for_bdist_rpm')
@@ -1477,10 +1545,8 @@ class vsc_setup(object):
 
         self.prepare_rpm(target)
         x = self.parse_target(target, urltemplate)
-
         setupfn(**x)
 
-
 # here for backwards compatibility
 SHARED_TARGET = _fvs('SHARED_TARGET').SHARED_TARGET
 
@@ -1493,14 +1559,25 @@ def action_target(package, *args, **kwargs):
     _fvs('action_target function')().action_target(package, *args, **kwargs)
 
 
+MAX_SETUPTOOLS_VERSION = '42.0'
+
 if __name__ == '__main__':
     """
     This main is the setup.py for vsc-install
     """
     install_requires = [
-        'setuptools',
+        # setuptools 42.0 changed easy_install to use pip if it's available,
+        # but vsc-install relies on the setuptools' behaviour of ignoring failing dependency installations and
+        # just continuing with the next entry in dependency_links
+        'setuptools < %s' % MAX_SETUPTOOLS_VERSION,
     ]
 
+    # mock 4.0 only supports Python 3+
+    if sys.version_info < (3, 0):
+        install_requires.append('mock < 4.0')
+    else:
+        install_requires.append('mock')
+
     PACKAGE = {
         'version': VERSION,
         'author': [sdw, ag, jt],
